/*
 * This file implements the siblings() method using the KVM interface
 * supported by BSD, and hopefully Macs also.  (BSD I've got.)
 */

#include <iostream>
#include <string>
using std::string;

#include <unistd.h>

#include <limits.h>
#include <fcntl.h>
#include <kvm.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/sysctl.h>

#include "find_siblings.h"

/*
 * This function builds and returns a vector of all sibling processes.
 */
std::vector<pid_t> siblings(bool safe /* = true */) 
{
	// Who are we?
	pid_t mypid = getpid();
	pid_t parpid = getppid();
	if(safe && parpid == 1)
		throw unsafe_sibling("Ooops! Loony seems to have been "
				     "inherited by init");

	// Open the KVM thingie.  This can be used to read a crash file,
	// in which case we're opening an actual file.  In the past, we
	// would read the kernel memory via a device file.  Now days it
	// seems to be a library that gets what you want from the kernel
	// using syscalls.
	char errbuf[_POSIX2_LINE_MAX];
	kvm_t *kvm = kvm_openfiles(NULL, NULL, NULL, KVM_NO_FILES, errbuf);
	if(kvm == NULL)
		throw std::runtime_error(string("kvm open fail: ") + errbuf);

	// Get the list of my processes.
	uid_t me = getuid();
	int cnt;
	struct kinfo_proc2 *procs =
		kvm_getproc2(kvm, KERN_PROC_RUID, me,
			     sizeof(struct kinfo_proc2), &cnt);
	if(procs == NULL) {
		string err(string("Process List Failed: ") + kvm_geterr(kvm));
		kvm_close(kvm);
		throw std::runtime_error(err);
	}

	// Go through and see what we got.
	std::vector<pid_t> ret;
	for(int i = 0; i < cnt; i++, procs++) {
		// We presume that our parent is also owned by use, and
		// in safe mode we need to examine it.
		if(safe && procs->p_pid == parpid) {
			// Fetch it's command line (main's arguments)
			char **args = kvm_getargv2(kvm, procs, 0);
			if(args == NULL) continue;
			
			// See if looks like a shell.  Strip off
			// any path (don't expect it), and any leading
			// dash which may be added to shells.
			string cmd = args[0];
			if(cmd[0] == '-') cmd.erase(0,1);
			int sl = cmd.find("/");
			if(sl != string::npos) cmd.erase(0,sl+1);
			if(cmd[0] == '-') cmd.erase(0,1);
			if(cmd == "bash" || cmd == "ksh" || cmd == "sh" ||
			   cmd == "csh" || cmd == "zsh")
				throw unsafe_sibling("Ooops! Loony seems to "
						     "have been run from "
						     "the command shell");
		}
		else if(procs->p_pid != mypid && procs->p_ppid == parpid) {
			ret.push_back(procs->p_pid);
		}
	}

	kvm_close(kvm);

	return ret;
}
