
#include "find_siblings.h"

// Unix headers.
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>

// Plain C header
#include <cctype>

#include <fstream>

// This is what we throw when we go boom.
#include <stdexcept>

using std::string;

/*
 * This reads a status file from /proc and returns the parent pid from
 * the file, or -1 if it fails.
 */
static pid_t extract_parent(string filename)
{
	// Mainly because the the file can disappear while we're
	// trying to read it, all sorts of bad stuff can happen.
	try {
		// Open the file.  If it fails, just figure the process exited
		// while we were reading the directory.
		std::ifstream pfile(filename.c_str());
		if(!pfile) return -1;

		// Read until we see PPid:, then the next thing will be the 
		// parent pid.
		string word;
		while(pfile >> word) {
			if(word == "PPid:") {
				// Next word should exist, and will be the 
				// parent pid.
				if(!(pfile >> word)) return -1;
				return atoi(word.c_str());
			}
		}
	} catch(...) {
	}

	return -1;
}

/*
 * This function builds and returns a vector of all sibling processes.
 */
std::vector<pid_t> siblings(bool safe /* = true */) 
{
	// Use the O/S calls to find our own pid and our parent's pid.
	pid_t mypid = getpid();
	pid_t mypar = getppid();

	// Now some safety checks.
	if(safe && mypar == 1)
		throw unsafe_sibling("Ooops! Loony seems to have been "
				     "inherited by the init process.  "
				     "Terminating.");

	// See if the parent process looks like it's a shell.
	{
		char target[1024];
		string exe = "/proc/" + std::to_string(mypar) + "/exe";
		auto size = readlink(exe.c_str(), target, sizeof target);
		if(size > 0) {
			string targ(target, size);
			if(safe && (targ == "/usr/bin/bash" ||
				    targ == "/bin/bash"))
				throw unsafe_sibling("Ooops! Loony seems to "
						     "have been run from "
						     "the command shell");
		}
	}

	// Now go through all the processes listed in the system /proc 
	// directory and add our siblings to ret.
	std::vector<pid_t> ret;
	DIR *dir = opendir("/proc");
	if(dir == NULL)
		throw std::runtime_error("Cannot open /proc: " +
					 string(strerror(errno)));
	string mypid_s = std::to_string(mypid);
	while(1) {
		// Read the next directory entry, stop if that's all.
		struct dirent *de = readdir(dir);
		if(de == 0) break;

		// See if the file name looks like it's a process number.
		// We'll just see if the first character is a digit.
		// If not, just skip this file.
		if(!isdigit(de->d_name[0])) continue;

		// Skip ourselves.
		string name = de->d_name;
		if(name == mypid_s) continue;

		// Extract the parent pid, and skip it if it's not the
		// same as ours.
		if(extract_parent("/proc/"+name+"/status") != mypar) continue;

		// Add to the list
		ret.push_back(atoi(name.c_str()));
	}
	closedir(dir);
	
	return ret;
}
