#include <iostream>
#include <vector>
#include <stdexcept>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>

#include <sys/wait.h>

#include "find_siblings.h"

/*
 * Print the list.
 */
void prpids(const std::vector<pid_t> &pids)
{
	for(auto p: pids)
		std::cout << p << " ";
	std::cout << std::endl;
}

void runtest()
{
	std::cout << "Running sibling test for " << getpid() << ", child of "
		  << getppid() << std::endl;
	
	try {
		auto full = siblings(false);
		std::cout << "Full sibling list: ";
		prpids(full);
	}
	catch(std::exception &e) {
		std::cout << "Hmmm.  unsafe call crashed: " << e.what()
			  << std::endl;
	}

	try {
		auto safe = siblings();
		std::cout << "Safe siblings: ";
		prpids(safe);
	}
	catch(unsafe_sibling &e) {
		std::cout << "Safe mode hits limit: " << e.what()
			  << std::endl;
	}
	catch(std::exception &e) {
		std::cout << "Hmmm.  safe call crashed in unexpected way: "
			  << e.what() << std::endl;
	}
}
int main()
{
	// Go.
	std::cout << "Process " << getpid() << " starting." << std::endl;
	
	// Run the test.
	std::cout << "Testing in the main process" << std::endl;
	runtest();
	std::cout << std::endl;

	// Make some bogus sibling processes for the lister to find.
	std::cout << "Creating test siblings." << std::endl;
	std::vector<pid_t> extras;
	for(int i = 0; i < 3; ++i) {
		pid_t kidpid = fork();
		if(kidpid < 0) {
			std::cout << "Fork failed: "
				  << strerror(errno) << std::endl;
		} else {
			if(kidpid == 0) { while(true); exit(0); }
			std::cout << "Started sibling " << kidpid
				  << std::endl;
			extras.push_back(kidpid);
		}
	}
	std::cout << std::endl;

	// Run the test in a child process.
	std::cout << "Creating child process for multi-child test"
		  << std::endl;
	pid_t kidpid = fork();
	if(kidpid < 0)
		std::cout << "Fork failed: " << strerror(errno)
			  << std::endl;
	else if(kidpid == 0) { runtest(); exit(0); }
	else {
		// Wait for the test child.
		int status;
		pid_t k = waitpid(kidpid, &status, 0);
		std::cout << "Multi-child test process " << k
			  << " exited." << std::endl;

		// Kill the siblings.
		for(auto p: extras) kill(p,SIGINT);

		// Wait the siblings.
		for(int i = 0; i < extras.size(); ++i) {
			pid_t k = wait(&status);
			std::cout << "Test sibling " << k << " exited."
				  << std::endl;
		}
	}
	std::cout << std::endl;

	// Try it once without siblings.
	std::cout << "Creating child process for only-child test"
		  << std::endl;
	kidpid = fork();
	if(kidpid < 0)
		std::cout << "Fork failed: " << strerror(errno)
			  << std::endl;
	else if(kidpid == 0) { runtest(); exit(0); }
	else {
		int status;
		pid_t k = wait(&status);
		std::cout << "Singleton test process " << k << " exited."
			  << std::endl;
	}

	std::cout << "Main process " << getpid() << " exiting." << std::endl;

	exit(0);
}
